// Class: PattRun
//
// Purpose: Java front-end gui wrapper for Bayesian Decomposition
//
// History: v0.7 5/03 DSC
//          v1.0 6/03 WFS
//          v1.1 6/04 WFS
//          v2.0 8/06 WFS
//          v2.2 11/06 WFS
//
// COPYRIGHT, 2006, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the 
// GNU General Public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along with this program;  if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA   02111-1307, USA.

import java.awt.Dimension;
import java.awt.BorderLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.ItemEvent;
import java.awt.event.ItemListener;

import java.io.BufferedReader;
import java.io.File;
import java.io.FilenameFilter;
import java.io.FileReader;
import java.io.IOException;

import java.net.JarURLConnection;
import java.net.URL;

import java.rmi.*;

import java.util.Enumeration;
import java.util.LinkedList;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import javax.swing.AbstractAction;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.SpringLayout;

import pattviewer.*;
import lib.*;
import rmi.*;
import textfields.*;
import viewer.*;

import static javax.swing.SpringLayout.*;

public class PattRun extends JFrame implements PattI{
    private static final long serialVersionUID = 2006022l;
    public static int version = 2006022;
    public static String copyright = "Copyright 2006, Fox Chase Cancer Center";

    boolean isRunning=false;    // true if run is invoked on this computer
    LinkedList<String> servers; // list of servers to run on
    static int port=1099;       // default port to run on server
    Library[] cVersion;         // list of available local c libraries
    Library[] libraries;

    GenpatExp gpe;              // output data from algorithm
    String[] nameView;          // names of the genes
    PattInput data;             // input data for algorithm

    PattI patti;                // Connected server

    //Start Swing Elements
    JComboBox serverCombo;  // Contains the names of the servers
    String addString = "Add...";
    DataField portField;    // Field for the user to enter the port number (default:1099)
    JTextArea serverStatus; // Displays information about the connected server
    JComboBox libraryCombo; // Contains the names of the available algorithms
    DataPane dPane;         // Displays the values of the PattInput fields
    CommandPane cPane;      // Displays the output text from the program

    JFileChooser loadingFile = new JFileChooser();

    /*
     * Construct the frame
     */
    public PattRun() {
	pattInit();
	serverChanged();

	boolean gotRegistry = false;
	while(!gotRegistry){
	    try{  //creates the rmi registry and does nothing if it already exists
		java.rmi.registry.LocateRegistry.createRegistry(port);
		gotRegistry = true;
	    }catch(java.rmi.server.ExportException ee){
		port++;
		gotRegistry = false;
	    }catch(RemoteException re){
		guiPrint(re);
	    }
	}
	try{  //creates the PattS to be used with rmi
	    Naming.rebind("//:"+Integer.toString(port)+"/PattS",new PattS(this));
	}catch(java.net.MalformedURLException mue){
	    guiPrint(mue);
	}catch(RemoteException re){
	    guiPrint(re);
	}
    }

    /*
     * GUI initialization
     */
    private void pattInit()  {
	this.setSize(800, 600);
	this.setDefaultCloseOperation(3);
	this.setTitle("PattRun");
	this.getContentPane().setLayout(new BorderLayout());
	this.setResizable(true);

	data = new PattInput();
	dPane = new DataPane(data);
	dPane.setDividerLocation(110);

	SpringLayout serverLayout = new SpringLayout();
	JPanel serverPanel = new JPanel();
	serverPanel.setLayout(serverLayout);

	JSplitPane leftSplitPane = new JSplitPane(JSplitPane.VERTICAL_SPLIT,true,dPane,serverPanel);
	leftSplitPane.setEnabled(true);
	leftSplitPane.setDividerSize(5);
	leftSplitPane.setDividerLocation(370);
	leftSplitPane.setResizeWeight(1);
	cPane = new CommandPane();

	JSplitPane split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,true,leftSplitPane,cPane);
	split.setEnabled(true);			//Split Pane
	split.setDividerSize(5);
	split.setDividerLocation(300);
	split.setResizeWeight(0);	
	this.getContentPane().add(split, BorderLayout.CENTER);

	JLabel sLabel = new JLabel("Server:");
	serverLayout.putConstraint(WEST, sLabel,5,WEST, serverPanel);
	serverLayout.putConstraint(NORTH, sLabel,5,NORTH, serverPanel);
	serverPanel.add(sLabel);

	servers=new LinkedList<String>();
	servers.add("local");
	try {     // read file PattServers.list for the names of the available servers
	    BufferedReader br=new BufferedReader(new FileReader(new File("PattServers.list")));
	    String tempStr=br.readLine();
	    while(tempStr!=null){
		servers.add(tempStr);
		tempStr=br.readLine();
	    }
	    br.close();
	}catch(IOException ioe) {
	    guiPrint("Could not read file: "+"PattServers.list\n");
	}
	guiPrint("servers:"); // print the names of the available servers
	for(int i=0;i<servers.size();i++){
	    guiPrint((String)servers.get(i));
	}
	guiPrint("");
	servers.add(addString);
	serverCombo = new JComboBox(servers.toArray());
	/*
	 * connect to new server when serverCombo is changed
	 */
	serverCombo.addItemListener(new ItemListener(){
		public void itemStateChanged(ItemEvent e){
		    if(e.getStateChange()==ItemEvent.SELECTED){
			serverChanged();
		    }
		}
	    });
	serverLayout.putConstraint(WEST, serverCombo, 5, EAST, sLabel);
	serverLayout.putConstraint(NORTH, serverCombo, 5,NORTH,serverPanel); 
	serverPanel.add(serverCombo);

	JLabel pLabel = new JLabel("Port:");
	serverLayout.putConstraint(WEST, pLabel,5,WEST,serverPanel);
	serverLayout.putConstraint(NORTH,pLabel,5,SOUTH,serverCombo);
	serverPanel.add(pLabel);

	portField=new DataField(10,port,new AbstractAction("connect"){
		public void actionPerformed(ActionEvent ae){
		    serverChanged();
		}
	    });
	serverLayout.putConstraint(WEST,portField,5,EAST,pLabel);
	serverLayout.putConstraint(NORTH,portField,5,SOUTH,serverCombo);
	serverPanel.add(portField);

	serverStatus = new JTextArea();
	serverStatus.setEditable(false);
	serverLayout.putConstraint(WEST, serverStatus, 5, WEST, serverPanel);
	serverLayout.putConstraint(NORTH, serverStatus, 5, SOUTH, portField);
	serverPanel.add(serverStatus);

	JLabel libraryLabel = new JLabel("Library:");
	serverLayout.putConstraint(WEST, libraryLabel, 5, WEST, serverPanel);
	serverLayout.putConstraint(NORTH,libraryLabel,5,SOUTH, serverStatus);
	serverPanel.add(libraryLabel);

	LinkedList<Library> libs = new LinkedList<Library>();
	URL url = PattRun.class.getResource("/lib");
	File rootFile = new File(url.getFile());
	if(rootFile.exists()){
	    String[] files = rootFile.list();
	    for(int i=0;i<files.length;i++){
		try{
		    if(files[i].endsWith(".class")){
			String  cname = files[i].substring(0,files[i].lastIndexOf("."));
			Library l = (Library)Class.forName("lib."+cname).newInstance();
			libs.add(l);
		    }
		}catch(Throwable t){}
	    }
	}
	try{
	    JarURLConnection conn = (JarURLConnection)url.openConnection();
	    String starts = conn.getEntryName();
	    JarFile jfile = conn.getJarFile();
	    Enumeration e = jfile.entries();
	    while(e.hasMoreElements()){
		ZipEntry entry = (ZipEntry)e.nextElement();
		String entryname = entry.getName();
		if(entryname.startsWith(starts)&&(entryname.lastIndexOf('/')<=starts.length())&&entryname.endsWith(".class")){
		    String classname = entryname.substring(0,entryname.lastIndexOf("."));
		    if(classname.startsWith(File.separator)){
			classname = classname.substring(1);
		    }
		    classname = classname.replace('/','.');
		    try{
			Library l = (Library)Class.forName(classname).newInstance();
			libs.add(l);
		    }catch(Throwable eee){}
		}
	    }
	}catch(Exception ee){}
	cVersion = new Library[libs.size()];

	guiPrint("available libraries:");
	for(int i=0;i<cVersion.length;i++){
	    cVersion[i]=libs.get(i);
	    cVersion[i].setParent(this);
	    guiPrint(cVersion[i]);
	}
	guiPrint("");
	libraryCombo = new JComboBox(cVersion);
	serverLayout.putConstraint(WEST, libraryCombo, 5,EAST, libraryLabel);
	serverLayout.putConstraint(NORTH,libraryCombo,5,SOUTH, serverStatus);
	libraryCombo.addItemListener(new ItemListener(){
		public void itemStateChanged(ItemEvent e){
		    if(e.getStateChange()==ItemEvent.SELECTED){
			libraryChanged();
		    }
		}
	    });
	serverPanel.add(libraryCombo);

	JPanel southPane = new JPanel();                              //south pane
	southPane.setPreferredSize(new Dimension(10, 50));
	this.getContentPane().add(southPane, BorderLayout.SOUTH);

        LinkedList<String> extensions=new LinkedList<String>();
        extensions.add("tab");
        extensions.add("txt");
	extensions.add("data");
	loadingFile.setFileFilter(new ImageFilter(extensions));
	
	//Creates 5 buttons with different action commands 
	String[] buttonNames = {"Load", "View", "Run","Clear"};	
	for (int i=0;i<buttonNames.length;i++) {
		JButton tempButton = new JButton();
        	tempButton.setPreferredSize(new Dimension(75,25));
   	     	tempButton.setActionCommand(Integer.toString(i));
		tempButton.setText(buttonNames[i]);
	   	tempButton.addActionListener(new java.awt.event.ActionListener() {
		   	public void actionPerformed(ActionEvent e) {
				tempButton_actionPerformed(e); 
			     }
		});
	     	southPane.add(tempButton);
	}
    }

    /*
     * Button Listener
     */	   
    void tempButton_actionPerformed(ActionEvent e) {
	switch(Integer.parseInt(((JButton)e.getSource()).getActionCommand())) {
	case 0:
	    load_actionPerformed(e);
	    break;
	case 2:
	    run_actionPerformed(e);
	    break;
	case 1:
	    view_actionPerformed(e);
	    break;
	case 3:
	    clear();
	    break;
	default:
	    System.exit(1);
	}
    }
   
    /*
     * Runs PattViewer with the newly created .gpe file(if 
     * patts was run on the loaded data)
     */
    void view_actionPerformed(ActionEvent e) {
	PattFrame pattview=new PattFrame();
	pattview.setVisible(true);
	try{
	    pattview.setData(gpe,nameView);
	}catch(NullPointerException npe){}
    }  
	
    /*
     * Load method.  Opens a JFileChooser to let user select .tab file 
     * to be loaded.  Method checks if the file is of valid type (not format)
     */	 
    void load_actionPerformed(ActionEvent e){
	if(JFileChooser.APPROVE_OPTION == loadingFile.showOpenDialog(this)){
	    String tag = loadingFile.getSelectedFile().getPath();
	    int cutExt = tag.lastIndexOf('.');
	    tabRead(tag);//this should probably throw an exception to be caught.
	}
    }

    /*
     * reads in all files sharing the rootname of the selected file and having
     * one of the accepted extensions: txt tab data unc cls ctrl ctr
     * see readme for file formats.
     * the names of the files read are printed on the commandpane
     */	
    void tabRead(String filename) {
	try{
	    LinkedList<String> files = new LinkedList<String>();
	    parsePattInput.parse(filename,data,files);
	    Object[] fArray = files.toArray();
	    for(int i=0;i<fArray.length;i++){
		guiPrint("read "+fArray[i].toString());
	    }
	    dPane.refresh();
	}catch(IOException ioe){
	    clear();
	    guiPrint("Error: Could not load "+filename);
	    guiPrint(ioe.toString());
	}
    }

    /*
     * Clears the gui of all information, clearing all
     * data fields and purging any data
     * stored in dataStr or acc arrays
     */	
    private void clear(){
	cPane.clear();
    }

    /*
     * action performed by the action listener when run is selected
     * creates a new PattThread and runs it.
     */
    void run_actionPerformed (ActionEvent e) {
	try{
	    if(libraryCombo.getSelectedIndex()==-1){
		guiPrint("No library selected");
	    }else if(!patti.isBusy()){
		File dir = new File(data.getDir());
		try{
		    guiPrint("running "+libraryCombo.getSelectedItem()+" on "+serverCombo.getSelectedItem());
		    if(!dir.exists()){
			guiPrint("ERROR: directory "+dir+" does not exist");
		    }else if(!dir.isDirectory()){
			guiPrint("ERROR: "+dir+" is not a directory");
		    }else{
			patti.run(libraryCombo.getSelectedIndex(),data,port);
		    }
		    data.randomizeIseed();
		    dPane.refresh();
		}catch(SecurityException se){
		    guiPrint("ERROR: access to directory "+dir+" is denied");
		}catch(Exception exception){
		    guiPrint("Error: JNI failure\n"+exception);
		}
	    }
	}catch(RemoteException re){
	    guiPrint("remote exception");
	}
    }

    private void libraryChanged(){
	dPane.setFields(libraries[libraryCombo.getSelectedIndex()].getFieldNames());
    }

    /*
     * Tries to connect to the server specified in serverCombo on the port set in portField
     */
    private void serverChanged(){
	String server = (String)serverCombo.getSelectedItem();
	int temp;
	if(portField.getText().equals("")){
	    temp = 0;
	}else{
	    temp= Integer.parseInt(portField.getText());
	}
	if(server.equals(addString)){
	    serverCombo.addItem("");
	    serverCombo.setSelectedItem("");
	    serverCombo.removeItemAt(serverCombo.getItemCount()-2);
	}else if(server.equals("")){
	    serverCombo.setEditable(true);
	}else{
	    serverCombo.setEditable(false);
	    serverCombo.removeItemAt(serverCombo.getItemCount()-1);
	    if(serverCombo.getSelectedIndex()==-1){
		serverCombo.addItem(server);
	    }
	    serverCombo.addItem(addString);
	    try{
		if(server.equals("local")){
		    patti=this;
		}else{
		    patti = (PattI) Naming.lookup ("//"+server+":"+temp+"/PattServ");
		}
		libraryCombo.removeAllItems();
		libraries = patti.getLibraries();
		for(int i=0;i<libraries.length;i++){
		    libraryCombo.addItem(libraries[i]);
		}
		setServerStatus();
		guiPrint("Connected to "+server+(server.equals("local")?"":" on port "+temp));
	    }catch(Exception e){
		guiPrint("Cannot connect to "+server+" on port "+temp);
		serverCombo.setSelectedIndex(0);
	    }
	}
    }

    /*
     * sets the serverStatus textPane to show the status of patti
     */
    private void setServerStatus(){
	try{
	    String[] status = patti.getStatus();
	    serverStatus.setText("");
	    String s = "os:"+status[0]+"\nversion:"+status[1]+"\nstatus:"+status[2];
	    serverStatus.setText(s);
	}catch(Exception e){
	    guiPrint(e);
	}
    }

    /*
     * Prints outputs to the CommandPane in the gui
     * Called by PattS.java when using rmi.
     */
    public void guiPrint(String text){
	//System.out.println(text);
	cPane.guiPrint(text);
    }

    /*
     * Convenience method calls tostring method of the object for the user.
     */
    public void guiPrint(Object text){
	guiPrint(text.toString());
    }

    /*
     * writes the GenpatExp object to a gpe file specified by filename
     */
    public void writePattOut(String filename, GenpatExp pattOut){
	try{
	    String f = parsePattOutput.writeGPE(filename,pattOut);
	    guiPrint("wrote "+f);
	    if(pattOut.getGeneNames().length==pattOut.getNa()){
		f = parsePattOutput.writeGNM(f,pattOut.getGeneNames());
		guiPrint("wrote "+f);
	    }
	    gpe = pattOut;
	    nameView = pattOut.getGeneNames();
	}catch(IOException ioe){
	    guiPrint("could not write "+filename);
	}
    }

    /*
     * returns whether an algorithm is running locally
     */
    public boolean isBusy(){
	return isRunning;
    }

    /*
     * creates and executes a thread on which cVersion[whichC] runs
     */
    public void run(int whichC, PattInput pattIn, int p){
	new PattThread(whichC,data).start();
    }

    /*
     * returns the availabel libraries that can run locally
     */
    public Library[] getLibraries(){
	return cVersion;
    }

    /*
     * returns the status of the connected server
     */
    public String[] getStatus()throws RemoteException{
	if(((String)serverCombo.getSelectedItem()).equals("local")){
	    String[] status = new String[3];
	    try{
		status[0]=System.getProperty("os.name");
		status[1]=System.getProperty("os.version");
	    }catch(Exception e){
		status[0]=e.toString();
		status[1]="";
	    }
	    status[2]=(isRunning)?"Busy":"Ready";
	    return status;
	}else{
	    setServerStatus();
	    return patti.getStatus();
	}
    }

    /*
    * runs Bayesian Decomposition either on the local host or on a
    * separate server using rmi.  The server list is in a file called
    * PattServers.list which is then saved into the LinkedList servers.
    * runs on a separate thread so the main thread can continue while
    * the process is being completed.
    */
    private class PattThread extends Thread{
	int whichC;
	PattInput pattIn;
	public PattThread(int wc, PattInput b){
		super();
		whichC = wc;
		pattIn = b;
	}
	public void run(){
	    isRunning = true;
	    setServerStatus();
	    GenpatExp output = cVersion[whichC].run(pattIn);
	    isRunning = false;
	    setServerStatus();
	}
    }

    /*
     * main method creates an instance and makes it visible
     */
    public static void main(String[] args) {
	PattRun frame = new PattRun();
	frame.setVisible(true);
    }
}//End of class PattRun
