// Class: PattServ
//
// Purpose: Server program for running PattRun with RMI
//
// History: v1.0 8/03 WFS
//          v2.1 8/06 WFS
//
// COPYRIGHT, 2003, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU Gerneral Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// NERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
// GNU General public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import java.io.File;
import java.io.FilenameFilter;

import java.net.JarURLConnection;
import java.net.URL;

import java.rmi.*;
import java.rmi.server.*;

import java.util.Enumeration;
import java.util.LinkedList;
import java.util.Properties;
import java.util.jar.JarFile;
import java.util.zip.ZipEntry;

import pattviewer.*;
import lib.*;
import rmi.*;

public class PattServ extends UnicastRemoteObject implements PattI {
    private static final long serialVersionUID = 2006021l;
    public static int version = 2006021;
    public static String copyright = "Copyright 2006, Fox Chase Cancer Center";

    PattI patti;
    boolean running=false;
    static int port;
    Library[] cVersion;

    /*
     * Constructor
     * reads in the list of classes in ./lib and tests which of them are valid.
     */
    public PattServ() throws RemoteException{
	LinkedList<Library> libs = new LinkedList<Library>();
	URL url = PattServ.class.getResource("/lib");
	File rootFile = new File(url.getFile());

	if(rootFile.exists()){
	    String[] files = rootFile.list();
	    for(int i=0;i<files.length;i++){
		try{
		    if(files[i].endsWith(".class")){
			String cname = files[i].substring(0,files[i].lastIndexOf("."));
			Library l = (Library)Class.forName("lib."+cname).newInstance();
			libs.add(l);
		    }
		}catch(Throwable t){}
	    }
	}
	try{
	    JarURLConnection conn = (JarURLConnection)url.openConnection();
	    String starts = conn.getEntryName();
	    JarFile jfile = conn.getJarFile();
	    Enumeration e = jfile.entries();
	    while(e.hasMoreElements()){
		ZipEntry entry = (ZipEntry)e.nextElement();
		String entryname = entry.getName();
		if(entryname.startsWith(starts)&&
		   (entryname.lastIndexOf("/")<=starts.length())&&
		   entryname.endsWith(".class")){
		    String classname = entryname.substring(0,entryname.lastIndexOf("."));
		    if(classname.startsWith("/")){
			classname = classname.substring(1);
		    }
		    classname = classname.replace('/','.');
		    try{
			Library l = (Library)Class.forName(classname).newInstance();
			libs.add(l);
		    }catch(Throwable t){}
		}
	    }
	}catch(Exception e){}
	cVersion = new Library[libs.size()];
	for(int i=0;i<cVersion.length;i++){
	    cVersion[i]=libs.get(i);
	    cVersion[i].setParent(this);
	}
    }

    /*
    * returns true if the program is busy running an algorithm
    */
    public boolean isBusy()throws RemoteException{
	return running;
    }

    /*
    * connects to the host PattRun program and then creates and executes a thread on which the library
    * cVersion[whichC] will run with input data pattIn
    */
    public void run(int whichC, PattInput pattIn,int p)throws RemoteException{
	if(!running){
	    try{
		patti = (PattI) Naming.lookup("rmi://"+getClientHost()+":"+Integer.toString(p)+"/PattS");
		System.out.println("Connected to "+getClientHost()+" on port "+Integer.toString(p));
		System.out.println("Busy");
	    }catch(Exception exception){
		System.out.println(exception);
	    }
	    new PattThread(whichC,pattIn).start();
	}
    }

    /*
    * sends the C output to PattS to be displayed on the client computer
    */
    public void guiPrint(String text)throws Exception{
	try{
            patti.guiPrint(text);
        }catch(Exception exception){
            System.out.println(text);
        }
    }

    /*
     * rmi method writes the output data from the library code to a file on the host computer
     */
    public void writePattOut(String filename, GenpatExp pattOut)throws RemoteException{
	try{
	    patti.writePattOut(filename,pattOut);
	}catch(java.rmi.ConnectException ce){
	    try{
		String f = parsePattOutput.writeGPE("temp/"+filename,pattOut);
		System.out.println("wrote "+f);
		if(pattOut.getGeneNames().length==pattOut.getNa()){
		    f = parsePattOutput.writeGNM(f,pattOut.getGeneNames());
		    System.out.println("wrote "+f);
		}
	    }catch(java.io.IOException ioe){
		System.out.println(ce);
	    }
	}
    }

    /*
     * returns the array of available algorithms that can run on this server.
     */
    public Library[] getLibraries(){
	return cVersion;
    }

    /*
     * returns an array of strings containing information about the server.
     */
    public String[] getStatus(){
	String[] status = new String[3];
	try{
	    status[0]=System.getProperty("os.name");
	    status[1]=System.getProperty("os.version");
	}catch(Exception e){
	    status[0]=e.toString();
	    status[1]="";
	}
	status[2]=running?"Busy":"Ready";
	return status;
    }

    /*
    * runs the C program on a separate thread so it can still communicate
    * with the client
    */
    private class PattThread extends Thread{
	int whichC;
	PattInput pattIn;

	/*
	 * wc is the index of cVersion which contains the library to run
	 */
	public PattThread(int wc,PattInput b){
	    super();
	    whichC = wc;
	    pattIn = b;
	}
	/*
	 *  runs the library cVersion[whichC] with input data pattIn
	 */
        public void run(){
	    try{
		running=true;
		patti.getStatus();
		GenpatExp output = cVersion[whichC].run(pattIn);
		running=false;
		patti.getStatus();
		System.out.println("Ready");
	    }catch(RemoteException re){}
	}
    }

    /*
    * creates the rmiregistry and the PattServ object on the server
    * runs on port 1099 or the next allowed port
    * prints "ready" once the server is ready to be called by PattRun
    */
    public static void main(String[] argv) {
	boolean gotRegistry = false;
	port = (argv.length==0)?1099:Integer.parseInt(argv[0]);
	while(!gotRegistry){
	    try{
		java.rmi.registry.LocateRegistry.createRegistry(port);
		gotRegistry = true;
	    }catch(java.rmi.server.ExportException exception){
		gotRegistry = false;
		port++;
	    }catch(java.rmi.RemoteException exception){
		System.out.println(exception);
	    }
	}
	System.out.println("running on port "+port);
	try{
	    Naming.rebind("//:"+Integer.toString(port)+"/PattServ",new PattServ());
	    System.out.println("Ready");
	}catch(Exception exception){
	    System.out.println(exception);
	}
    }
}// end of class PattServ