// Class: Library
//
// Purpose: wrapper class for libraries used in PattRun. There are two abstract functions that 
//          need to be overwritten: run which is the response to the run button in PattRun, and
//          getLib which returns the name of the C library. Everything else is designed to aid
//          in wrapping in JNI; if this is not needed most functions can be overwritten or ignored.
//
// History v1.0 7/06 WFS
//
// COPYRIGHT, 2006, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along with this program;  if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA   02111-1307, USA.

package lib;

import java.io.File;
import java.io.IOException;
import java.io.Serializable;

import java.rmi.*;
import java.rmi.server.*;

import pattviewer.*;
import rmi.PattI;

public abstract class Library implements Serializable{
    private static final long SerialVersionID = 200601l;

    PattI parent;

    /*
     * constructor. By default it is set to load a C library for JNI. If the library
     * does not use JNI, either catch the UnsatisfiedLinkError, or overwrite this constructor.
     */
    public Library()throws SecurityException, IOException, UnsatisfiedLinkError{
	String name = System.mapLibraryName(getLib());
	File lib = new File("lib");
	System.load(lib.getCanonicalPath()+File.separator+name);
    }

    /*
     * sets the PattI object that called this method so it can call parent methods
     */
    public void setParent(PattI p){
	parent = p;
    }

    /*
     * overrides Object toString method to return the name of the C file
     */
    public String toString(){
	return getName();
    }

    /*
     * returns the names of the applicable fields.
     * should be overwritten to include only desired field names.
     */
    public String[] getFieldNames(){
	return new String[]{"tag","dir","ENSEMBLE","Niter","Nshow","Iseed","AlphaA",
			    "AlphaF","lineMeasure","Nsol","Na","Nf","ptwidth",
			    "data","acc","additive","multiplicative","xLabels",
			    "geneNames","SPLIT_portion","NfeetA","NfootA",
			    "NlistA","Wtype","Nfeet","feetsizes","uncertFlag",
			    "StartRank","EndRank","Nchains"};
    }

    /*
     * returns the name of the C program called
     */
    public String getName(){
	return getLib();
    }

    /*
     * tests whether the PattInput object is valid input for the c code
     * this should be overwritten for different libraries which have
     * different constraints on their inputs. this code checks for some 
     * common constraints and can be called using Library.validateInput(input)
     */
    public static void validateInput(PattInput input)throws Exception{
	int Na = input.getNa();
	int Nf = input.getNf();
	int Nsol = input.getNsol();
	if(Na<=0)
	    throw new Exception("Na = "+Na+" . must be > 0");
	if(Nf<=0)
	    throw new Exception("Nf = "+Nf+" . must be > 0");
	if(input.getData().length!=Na*Nf)
	    throw new Exception("Data is of length "+input.getData().length+
				" . must be of length Na * Nf ("+(Na*Nf)+")");
	if(input.getIseed()<0)
	    throw new Exception("Iseed = "+input.getIseed()+" . must be >= 0");
	if(input.getNiter()<1)
	    throw new Exception("Niter = "+input.getNiter()+" . must be > 0");
	if(input.getAlphaA()==0)
	    throw new Exception("AlphaA cannot equal 0");
	if(input.getAlphaF()==0)
	    throw new Exception("AlphaF cannot equal 0");
	if(input.getENSEMBLE()!=1)
	    throw new Exception("ENSEMBLE must equal 1");
	for(int i=0;i<input.getAcc().length;i++){
	    if(input.getUncertFlag()==1&&input.getAcc()[i]==0)
		throw new Exception("Acc["+i+"] = 0 . 0 accuracy results in divide by zero error");
	}
	if(input.getUncertFlag()==1&&input.getAcc().length!=Na*Nf)
	    throw new Exception("Acc is of length "+input.getAcc().length+
				" . must be of length Na * Nf ("+(Na*Nf)+")\n"+
				"If no files contain uncertainty data, set uncertFlag to 0 and use "+
				"additive and multiplicative to set uncertainty.");
    }

    /*
     * writes sends the text output from the program to the parent
     */
    public void guiPrint(String output){
	try{
	    parent.guiPrint(output);
	}catch(Exception e){
	    System.out.println(e);
	}
    }

    /*
     * sends the GenpatExp object to the parent to be written to a file
     */
    public void writePattOut(String filename, GenpatExp pattOut){
	try{
	    parent.writePattOut(filename,pattOut);
	}catch(Exception e){
	    System.out.println(e);
	}
    }

    /*
     * abstract method which runs the native C method
     */
    public abstract GenpatExp run(PattInput pattIn);

    /*
     * abstract method which returns the name of the C library for JNI.
     * If this library does not use JNI, overwrite to return desired value of toString()
     */
    public abstract String getLib();
}
