// Class: parsePattInput
//
// Purpose: Parses data input a serialized PattInput object
//
// History v1.0 8/06 WFS
//
// COPYRIGHT, 2006, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 0211-1307, USA.

import java.io.*;

import java.util.LinkedList;
import java.util.ListIterator;

import pattviewer.*;

public abstract class parsePattInput {
    public static int version = 200601;
    public static String copyright = "Copyright 2006, Fox Chase Cancer Center";

    /*
     * tries to read in data from any file sharing the same rootname as filename, 
     * adds the filename to names, and returns the data in a PattInput object
     */
    public static PattInput parse(String filename, LinkedList<String> names)throws IOException{
	PattInput pattIn = new PattInput();
	int lastIndex = filename.lastIndexOf(".");
	String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	try{
	    parseUnc(rootname+".unc",pattIn);
	    names.add(rootname+".unc");
	}catch(FileNotFoundException fnfe){}
	try{
	    parseTab(rootname+".tab",pattIn);
	    names.add(rootname+".tab");
	}catch(FileNotFoundException fnfe){
	    try{
		parseTab(rootname+".txt",pattIn);
		names.add(rootname+".txt");
	    }catch(FileNotFoundException fnfe2){
		try{
		    parseData(rootname+".data",pattIn);
		    names.add(rootname+".data");
		}catch(FileNotFoundException fnfe3){}
	    }
	}
	try{
	    parseCls(rootname+".cls",pattIn);
	    names.add(rootname+".cls");
	}catch(FileNotFoundException fnfe){}
	try{
	    parseCtrl(rootname+".ctrl",pattIn);
	    names.add(rootname+".ctrl");
	}catch(FileNotFoundException fnfe){
	    setCtrlDefaults(pattIn);
	}
	try{
	    parseLink(rootname+".link",pattIn);
	    names.add(rootname+".link");
	}catch(FileNotFoundException fnfe){}
	try{
	    parseCtr(rootname+".ctr",pattIn);
	    names.add(rootname+".ctr");
	}catch(FileNotFoundException fnfe){}
	return pattIn;
    }

    /*
     * calls parse above, and discards the names of the files
     */
    public static PattInput parse(String filename)throws IOException, FileNotFoundException{
	return parse(filename,new LinkedList<String>());
    }

    /*
     * Convenience method
     * reads the data from the various input files into a serialized PattInput variable.
     * if different versions exist (eg filename.tab and filename.txt) or if the files
     * are not named using the assumed convention, the read methods should be called
     * individually.
     */
    public static void parse(String filename, PattInput pattIn)throws IOException,FileNotFoundException{
	PattInput temp = parse(filename);
	pattIn.clear();
	pattIn.set(temp);
    }

    /*
     * calls parse above, and stores the data from the PattInput object in pattIn
     */
    public static void parse(String filename, 
			     PattInput pattIn,
			     LinkedList<String> names)throws IOException, FileNotFoundException{
	pattIn.set(parse(filename,names));
    }

    /*
     * Conveniance method
     * determines what kind of file it is by its extension and then calls the
     * appropriate read method.
     */
    public static void parseFile(String filename,
				 PattInput pattIn)throws IOException, FileNotFoundException{
	String extension = filename.substring(filename.lastIndexOf("."));
	if(extension.equals(".txt")||extension.equals(".tab")){
	    parseTab(filename, pattIn);
	}else if(extension.equals(".cls")){
	    parseCls(filename, pattIn);
	}else if(extension.equals(".unc")){
	    parseUnc(filename, pattIn);
	}else if(extension.equals(".ctrl")){
	    parseCtrl(filename, pattIn);
	}else if(extension.equals(".ctr")){
	    parseCtr(filename, pattIn);
	}else if(extension.equals(".data")){
	    parseData(filename, pattIn);
	}else{
	    throw new FileNotFoundException("File extension not recognized");
	}
    }

    /*
     * reads in data from files with .tab or .txt extension
     */
    public static void parseTab(String filename,
				PattInput pattIn) throws IOException,FileNotFoundException{
	int Na;
	int Nf;
	int Nsol;
	int ptwidth;
	String[] xLabels;
	String[] geneNames;
	double[] data;
	String tag;
	String dir;

	BufferedReader tabBR = new BufferedReader(new FileReader(filename));
	try{
	    int slashIndex = filename.lastIndexOf(File.separator);
	    int dotIndex = filename.lastIndexOf(".");
	    tag = (dotIndex>slashIndex)?filename.substring(slashIndex+1,dotIndex):
		filename.substring(slashIndex+1);
	    dir = filename.substring(0,slashIndex+1);

	    String[] line;
	    line = tabBR.readLine().split("\\t");
	    xLabels = new String[line.length-1];
	    for(int i=0;i<xLabels.length;i++){
		xLabels[i]=line[i+1];
	    }
	    
	    Nf = xLabels.length;
	    Na=0;
	    LinkedList<Double> tempStorage = new LinkedList<Double>();
	    LinkedList<String> tempNames = new LinkedList<String>();
	    
	    while(true){
		try{
		    line=tabBR.readLine().split("\\t");
		    if(line.length!=Nf+1){
			break;
		    }
		    tempNames.add(line[0]);
		    for(int i=1;i<line.length;i++){
			tempStorage.add(Double.valueOf(line[i]));
		    }
		    Na++;
		}catch(NullPointerException npe){
		    break;
		}
	    }

	    geneNames=new String[Na];
	    tempNames.toArray(geneNames);
	    data = new double[Na*Nf];
	    ListIterator<Double> iter = tempStorage.listIterator(0);
	    for(int i=0;i<data.length;i++){
		data[i]=iter.next();
	    }

	    Nsol = 5;     //default
	    ptwidth = 1;  //default
	}catch(Exception nfe){
	    throw new IOException("incorrect format: .tab file\n"+nfe);
	}	    
	tabBR.close();

	pattIn.setNa(Na);
	pattIn.setNf(Nf);
	pattIn.setNsol(Nsol);
	pattIn.setPtwidth(ptwidth);
	pattIn.setXLabels(xLabels);
	pattIn.setGeneNames(geneNames);
	pattIn.setData(data);
	pattIn.setTag(tag);
	pattIn.setDir(dir);
    }

    /*
     * reads in data from files with .data extension
     * format:
     *
     * file name
     * description
     * Nsol Na Nf
     * sigmaFlag
     * sigmaValue
     * ptwidth
     * data [acc]
     */
    public static void parseData(String filename, PattInput pattIn)throws IOException,FileNotFoundException{
	int Na;
	int Nf;
	int Nsol;
	double ptwidth;
	double[] data;
	double[] acc;
	String tag;
	String dir;

	BufferedReader dataBR = new BufferedReader(new FileReader(filename));
	try{
	    dir = filename.substring(0,filename.lastIndexOf(File.separator)+1);

	    String[] line;
	    tag = dataBR.readLine();//file name
	    dataBR.readLine();//description
	    line = dataBR.readLine().trim().split("\\s++");//Nsol Na Nf
	    try{
		Nsol = Integer.parseInt(line[0]);
	    }catch(NumberFormatException nfe){
		Nsol = 5; // set default if Nsol is not provided
	    }
	    Na = Integer.parseInt(line[1]);
	    Nf = Integer.parseInt(line[2]);
	    line = dataBR.readLine().trim().split("\\s++");//sigmaFlag (=1 if no acc)
	    int sigmaFlag = Integer.parseInt(line[0]);
	    line = dataBR.readLine().trim().split("\\s++");//sigmaValue
	    double sigmaValue = Double.parseDouble(line[0]);
	    line = dataBR.readLine().trim().split("\\s++");//ptwidth
	    ptwidth = Double.parseDouble(line[0]);
	    
	    data = new double[Na*Nf];
	    acc = new double[Na*Nf];
	    for(int i=0;i<Na*Nf;i++){
		line = dataBR.readLine().trim().split("\\s++");//data [acc]
		data[i]=Double.parseDouble(line[0]);
		acc[i]=(sigmaFlag==1)?1d/sigmaValue:Double.parseDouble(line[1]);
	    }

	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .data file\n"+nfe);
	}
	dataBR.close();

	pattIn.setUncertFlag(1);
	pattIn.setNa(Na);
	pattIn.setNf(Nf);
	pattIn.setNsol(Nsol);
	pattIn.setPtwidth(ptwidth);
	pattIn.setData(data);
	pattIn.setAcc(acc);
	pattIn.setTag(tag);
	pattIn.setDir(dir);
    }

    /*
     * reads in data from files with .link extension
     * format:
     *
     * description
     * SPLIT_portion
     * Wtype
     * NfeetA
     * NfootA
     * NlistA
     */
    public static void parseLink(String filename, PattInput pattIn)throws IOException,FileNotFoundException{
	double SPLIT_portion;
	int NfeetA;
	int[] NfootA;
	int[][] NlistA;
	int Wtype;

	BufferedReader linkBR = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;
	    linkBR.readLine();//description
	    line = linkBR.readLine().trim().split("\\s++"); // SPLIT_portion
	    SPLIT_portion = Double.parseDouble(line[0]);
	    line = linkBR.readLine().trim().split("\\s++"); // Wtype
	    Wtype = Integer.parseInt(line[0]);
	    line = linkBR.readLine().trim().split("\\s++"); // Nfeet NfootA
	    NfeetA = Integer.parseInt(line[0]);
	    NfootA = new int[NfeetA];
	    int maxFoot = 0;
	    for(int i=0;i<NfeetA;i++){
		NfootA[i]=Integer.parseInt(line[i+1]);
		maxFoot = Math.max(maxFoot,NfootA[i]);
	    }
	    NlistA = new int[NfeetA][maxFoot];
	    for(int i=0;i<NfeetA;i++){
		for(int j=0;j<NfootA[i];j++){
		    line = linkBR.readLine().trim().split("\\s++");
		    int dataPt = Integer.parseInt(line[0]);
		    if(dataPt>pattIn.getNa()){
			NfeetA = 0;
			SPLIT_portion = 0;
			System.out.println("There is no such gene: regulon "+i+", gene "+dataPt);
			return;
		    }
		    NlistA[i][j]=dataPt;
		}
	    }
	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .link file\n"+nfe);
	}
	linkBR.close();

	pattIn.setSPLIT_portion(SPLIT_portion);
	pattIn.setNfeetA(NfeetA);
	pattIn.setNfootA(NfootA);
	pattIn.setNlistA(NlistA);
	pattIn.setWtype(Wtype);
    }

    /*
     * reads in data from files with .cls extension
     * format:
     *
     * Number of classes   Nfeet    Elements per class      feetsizes
     */
    public static void parseCls(String filename, PattInput pattIn)throws IOException,FileNotFoundException{
	int Nfeet;
	int[] feetsizes;

	BufferedReader clsBR = new BufferedReader(new FileReader(filename));
	try{
	    String[] cls = clsBR.readLine().split("\\t");
	    Nfeet = Integer.parseInt(cls[1]);
	    feetsizes = new int[Nfeet];
	    for(int i=0;i<feetsizes.length;i++){
		feetsizes[i]=Integer.parseInt(cls[i+3]);
	    }
	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .cls file\n"+nfe);
	}
	clsBR.close();

	pattIn.setNfeet(Nfeet);
	pattIn.setFeetsizes(feetsizes);
    }

    /*
     * reads uncertainties from a .unc file. Must be called after data from a .tab, .txt,
     * or .data file is read in.
     */
    public static void parseUnc(String filename, PattInput pattIn)throws IOException,FileNotFoundException{
	int Na;
	int Nf;
	String[] xLabels;
	String[] geneNames;
	double[] acc;

	BufferedReader tabBR = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;
	    line = tabBR.readLine().split("\\t");
	    xLabels = new String[line.length-1];
	    for(int i=0;i<xLabels.length;i++){
		xLabels[i]=line[i+1];
	    }
	    
	    Nf = xLabels.length;
	    Na=0;
	    LinkedList<Double> tempStorage = new LinkedList<Double>();
	    LinkedList<String> tempNames = new LinkedList<String>();
	    
	    while(true){
		try{
		    line=tabBR.readLine().split("\\t");
		    if(line.length<Nf+1){
			break;
		    }
		    tempNames.add(line[0]);
		    for(int i=1;i<line.length;i++){
			tempStorage.add(Double.valueOf(line[i]));
		    }
		    Na++;
		}catch(NullPointerException npe){
		    break;
		}
	    }

	    geneNames=new String[Na];
	    tempNames.toArray(geneNames);
	    acc = new double[Na*Nf];
	    ListIterator<Double> iter = tempStorage.listIterator(0);
	    for(int i=0;i<acc.length;i++){
		acc[i]=iter.next();
	    }
	}catch(Exception nfe){
	    throw new IOException("incorrect format: .unc file\n"+nfe);
	}	    
	tabBR.close();

	pattIn.setUncertFlag(1);
	pattIn.setNa(Na);
	pattIn.setNf(Nf);
	pattIn.setXLabels(xLabels);
	pattIn.setGeneNames(geneNames);
	pattIn.setAcc(acc);
    }

    /*
     * reads in data from files with .ctrl extension.
     * format:
     *
     * file name
     * description
     * ENSEMBLE
     * Niter
     * Nshow
     * Iseed
     * AlphaA
     * AlphaF
     * lineMeasure
     */
    public static void parseCtrl(String filename, PattInput pattIn)throws IOException, FileNotFoundException{
	int ENSEMBLE;
	int Niter;
	int Nshow;
	Long Iseed;
	double AlphaA;
	double AlphaF;
	double lineMeasure;

	BufferedReader ctrlBR = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;
	    line = ctrlBR.readLine().split("\\s++"); // file name ignored
	    line = ctrlBR.readLine().split("\\s++");  // description ignored
	    line = ctrlBR.readLine().split("\\s++"); // ENSEMBLE
	    ENSEMBLE = Integer.parseInt(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // Niter
	    Niter = Integer.parseInt(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // Nshow
	    Nshow = Integer.parseInt(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // Iseed
	    Iseed = Long.parseLong(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // AlphaA
	    AlphaA = Double.parseDouble(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // AlphaF
	    AlphaF = Double.parseDouble(line[0]);
	    line = ctrlBR.readLine().split("\\s++"); // lineMeasure
	    lineMeasure = Double.parseDouble(line[0]);
	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .ctrl file"+nfe);
	}
	ctrlBR.close();

	pattIn.setENSEMBLE(ENSEMBLE);
	pattIn.setNiter(Niter);
	pattIn.setNshow(Nshow);
	pattIn.setIseed((int)(Iseed&0x7FFFFFFFl));
	pattIn.setAlphaA(AlphaA);
	pattIn.setAlphaF(AlphaF);
	pattIn.setLineMeasure(lineMeasure);
    }

    /*
     * sets the default values for the fields that are usually set by the .ctrl file when
     * no such file is found
     */
    public static void setCtrlDefaults(PattInput pattIn){
	pattIn.setENSEMBLE(1);
	pattIn.setNiter(1000);
	pattIn.setNshow(500);
	pattIn.randomizeIseed();
	pattIn.setAlphaA((double)pattIn.getNa()*pattIn.getNsol());
	pattIn.setAlphaF((double)pattIn.getNf()*pattIn.getNsol());
	pattIn.setLineMeasure(1);
    }

    /*
     * reads in data from files with .ctr extension
     * format:
     *
     * Na         Nf       UncertFlag
     * startRank  endRank  Nchains      Niter
     */
    public static void parseCtr(String filename, PattInput pattIn)throws IOException, FileNotFoundException{
	int Na;
	int Nf;
	int uncertFlag;
	int startRank;
	int endRank;
	int Nchains;
	int Niter;

	BufferedReader ctrBR = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;
	    line = ctrBR.readLine().split("\\s++");
	    Na = Integer.parseInt(line[0]);
	    Nf = Integer.parseInt(line[1]);
	    uncertFlag = Integer.parseInt(line[2]);
	    line = ctrBR.readLine().split("\\s++");
	    startRank = Integer.parseInt(line[0]);
	    endRank = Integer.parseInt(line[1]);
	    Nchains = Integer.parseInt(line[2]);
	    Niter = Integer.parseInt(line[3]);
	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .ctr file\n"+nfe);
	}
	ctrBR.close();

	pattIn.setNa(Na);
	pattIn.setNf(Nf);
	pattIn.setUncertFlag(uncertFlag);
	pattIn.setStartRank(startRank);
	pattIn.setEndRank(endRank);
	pattIn.setNchains(Nchains);
	pattIn.setNiter(Niter);
    }

    /*
     * writes a .gpe file containing the serialized PattInput object pattIn
     */
    public static String writeGPE(String filename, PattInput pattIn)throws IOException{
	new File(filename.substring(0,filename.lastIndexOf(File.separator))).mkdirs();
	String finalName = "";
	try{
	    int lastIndex = filename.lastIndexOf(".");
	    String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	    int count = 1;
	    String tempname = rootname;
	    while(new File(tempname+".gpe").exists()){
		tempname = rootname+"_"+Integer.toString(count);
		count++;
	    }
	    ObjectOutputStream out = 
		new ObjectOutputStream(new FileOutputStream(new File(filename+".pge")));
	    out.writeObject(pattIn);
	    out.close();
	}catch(Exception e){
	    throw new IOException("error writing .gpe file\n"+e);
	}
	return finalName;
    }

    public static void main(String argv[]){
	try{
	    writeGPE(argv[0],parse(argv[0]));
	}catch(Exception e){
	    System.out.println(e);
	    System.exit(1);
	}
    }
}
