// Class: parsePattOutput
//
// Purpose: Parses data into a serialized GenpatExp object
//
// History: v0.1 6/01 MFO
//          v0.2 8/01 WFS
//          v0.3 6/02 DSC
//          v1.0 7/03 WFS
//          v2.0 8/06 WFS
//
// COPYRIGHT, 2006, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

import java.io.*;

import pattviewer.*;

public abstract class parsePattOutput {
    public static int version = 200601;
    public static String copyright = "Copyright 2006, Fox Chase Cancer Center";

    /*
     * static method creates a GenpatExp object from the data in the .inf and .out files
     * that share the same root name as filename
     */
    public static GenpatExp parse(String filename)throws FileNotFoundException, IOException{
	GenpatExp pattOut = new GenpatExp();
	int lastIndex = filename.lastIndexOf(".");
	String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	parseOut(rootname+".out",pattOut);
	parseInf(rootname+".inf",pattOut);
	return pattOut;
    }

    /*
     * static method reads in data from the .inf and .out files with the same root name as
     * filename and stores it in pattOut
     */
    public static void parse(String filename, GenpatExp pattOut)throws FileNotFoundException, IOException{
	pattOut.set(parse(filename));
    }

    /*
     * reads in data from inf file filename and stores it in pattOut
     */
    public static void parseInf(String filename, GenpatExp pattOut)throws FileNotFoundException, IOException{
	String tag;
	int Nsol;
	int Na;
	int Nf;
	String date;
	int Nfoot;
	double[] footprint;
	double[][] Distribution;
	double[][] DSigma;
	double[][] Pattern;
	double[][] PSigma;

 	BufferedReader br = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;

	    line = br.readLine().trim().split("\\s++"); // tag
	    tag = line[0];

	    line = br.readLine().trim().split("\\s++"); // Nsol Na Nf
	    Nsol = Integer.parseInt(line[0]);
	    Na = Integer.parseInt(line[1]);
	    Nf = Integer.parseInt(line[2]);

	    date = br.readLine();

	    line = br.readLine().trim().split("\\s++"); // Nfoot
	    Nfoot = Integer.parseInt(line[0]);	 

	    footprint = new double[Nfoot];
	    for(int i=0; i < Nfoot; i++){
		line = br.readLine().trim().split("\\s++"); // footprint
		footprint[i] = Double.parseDouble(line[0]);
	    }
	    
	    Distribution = new double[Na][Nsol];
	    DSigma = new double[Na][Nsol];
	    for(int i = 0; i < Nsol; i++) {
		for(int j = 0; j < Na; j++)  {
		    line = br.readLine().trim().split("\\s++"); // meanA sdA
		    Distribution[j][i] = Double.parseDouble(line[0]);
		    DSigma[j][i] = Double.parseDouble(line[1]);
		}
	    }
	    
	    Pattern = new double[Nsol][Nf];
	    PSigma = new double[Nsol][Nf];
	    for(int i = 0; i < Nsol; i++) {
		for(int j = 0; j < Nf; j++) { 
		    line = br.readLine().trim().split("\\s++"); // meanF sdF
		    Pattern[i][j] = Double.parseDouble(line[0]);
		    PSigma[i][j] = Double.parseDouble(line[1]);
		}
	    }

	}catch(Exception nfe){
	    throw new IOException("Incorrect format: .out file\n"+nfe);
	}
	br.close();

	pattOut.setTag(tag);
	pattOut.setNsol(Nsol);
	pattOut.setNa(Na);
	pattOut.setNf(Nf);
	pattOut.setDate(date);
	pattOut.setNfoot(Nfoot);
	pattOut.setFootprint(footprint);
	pattOut.setDistribution(Distribution);
	pattOut.setDistSD(DSigma);
	pattOut.setPattern(Pattern);
	pattOut.setPattSD(PSigma);
    }

    /*
     * reads in data from out file filename and stores it in pattOut
     */
    public static void parseOut(String filename, GenpatExp pattOut)throws FileNotFoundException, IOException{
	String tag;
	double Chisq;
	double AtomsA;
	double AtomsF;
	double Evidence;
	int ENSEMBLE;
	int Niter;
	int Nshow;
	int Iseed;
	int SeedP = 0;
	int Nsol;
	int Na;
	int Nf;
	double lineMeasure;
	double ptwidth;
	double[][] data;
	double[][] mock;
	double[][] noise;

	BufferedReader outbr = new BufferedReader(new FileReader(filename));
	try{
	    String[] line;
	    
	    line = outbr.readLine().trim().split("\\s++"); // tag
	    tag = line[0];

	    line = outbr.readLine().trim().split("\\s++"); // Chisq AtomsA AtomsF Evidence
	    Chisq = Double.parseDouble(line[0]);
	    AtomsA = Double.parseDouble(line[1]);
	    AtomsF = Double.parseDouble(line[2]);
	    Evidence = Double.parseDouble(line[3]);

	    line = outbr.readLine().trim().split("\\s++"); // ENSEMBLE Niter Nshow Iseed
	    ENSEMBLE = Integer.parseInt(line[0]);
	    Niter = Integer.parseInt(line[1]);
	    Nshow = Integer.parseInt(line[2]);
	    Iseed = Integer.parseInt(line[3]);
	    try{
		SeedP = Integer.parseInt(line[4]);
	    }catch(IndexOutOfBoundsException ioobe){
	    }catch(NumberFormatException nfe){}

	    line = outbr.readLine().trim().split("\\s++"); // Nsol Na Nf
	    Nsol = Integer.parseInt(line[0]);
	    Na = Integer.parseInt(line[1]);
	    Nf = Integer.parseInt(line[2]);

	    line = outbr.readLine().trim().split("\\s++"); // lineMeasure ptwidth
	    lineMeasure = Double.parseDouble(line[0]);
	    ptwidth = Double.parseDouble(line[1]);

	    line = outbr.readLine().trim().split("\\s++"); // linemeasure ptwidth Nf
	    data = new double[Na][Nf];
	    mock = new double[Na][Nf];
	    noise = new double[Na][Nf];
	    for(int i = 0; i < Na; i++) {
		for(int j = 0; j < Nf; j++) {
		    line = outbr.readLine().trim().split("\\s++"); // data mock acc
		    data[i][j] = Double.parseDouble(line[0]);
		    mock[i][j] = Double.parseDouble(line[1]);
		    noise[i][j] = Double.parseDouble(line[2]);
		}
	    }
	}catch(Exception ioe){
	    throw new IOException("Incorrect format: .out file\n"+ioe);
	}
	outbr.close();

	pattOut.setTag(tag);
	pattOut.setChisq(Chisq);
	pattOut.setAtomsA(AtomsA);
	pattOut.setAtomsF(AtomsF);
	pattOut.setEvidence(Evidence);
	pattOut.setENSEMBLE(ENSEMBLE);
	pattOut.setNiter(Niter);
	pattOut.setNshow(Nshow);
	pattOut.setIseed(Iseed);
	pattOut.setSeedP(SeedP);
	pattOut.setNsol(Nsol);
	pattOut.setNa(Na);
	pattOut.setNf(Nf);
	pattOut.setLineMeasure(lineMeasure);
	pattOut.setPtwidth(ptwidth);
	pattOut.setData(data);
	pattOut.setMock(mock);
	pattOut.setNoise(noise);
    }

    /*
     * writes the GenpatExp object into a serialized gpe file
     */
    public static String writeGPE(String filename, GenpatExp pattOut)throws IOException{
	String finalName = "";
	try{
	    filename = filename.replace("/",File.separator);
	    new File(filename.substring(0,filename.lastIndexOf(File.separator))).mkdirs();
	    int lastIndex = filename.lastIndexOf(".");
	    String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	    int count = 1;
	    String tempname=rootname;
	    while(new File(tempname+".gpe").exists()){
		tempname = rootname+"_"+Integer.toString(count);
		count++;
	    }
	    ObjectOutputStream oos = 
		new ObjectOutputStream(new FileOutputStream(new File(tempname+".gpe")));
	    finalName = tempname+".gpe";
	    oos.writeObject(pattOut);
	    oos.close();
	}catch(Exception e){
	    throw new IOException("error writing .gpe file\n"+e);
	}
	return finalName;
    }

    /*
     * writes the genenames into a gnm text file
     */
    public static String writeGNM(String filename, String[] geneNames)throws IOException{
	String finalName = "";
	try{
	    new File(filename.substring(0,filename.lastIndexOf(File.separator))).mkdirs();
	    int lastIndex = filename.lastIndexOf(".");
	    String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	    int count = 1;
	    String tempname=rootname;
	    while(new File(tempname+".gnm").exists()){
		tempname = rootname+"_"+Integer.toString(count);
		count++;
	    }
	    finalName = tempname+".gnm";
	    BufferedWriter gnmbw = new BufferedWriter(new FileWriter(finalName));
	    for(int i=0;i<geneNames.length;i++){
		gnmbw.write(geneNames[i]+"\n");
	    }
	    gnmbw.close();
	}catch(Exception e){
	    throw new IOException("error writing .gnm file\n"+e);
	}
	return finalName;
    }

    /*
     * reads the GenpatExp object from a file and then unparses
     */
    public static String unparse(String pattFile)throws IOException{
	String tempname;
	try{
	    ObjectInputStream ois = 
		new ObjectInputStream(new BufferedInputStream(new FileInputStream(pattFile)));
	    tempname = unparse(pattFile,(GenpatExp)ois.readObject());
	    ois.close();
	}catch(ClassNotFoundException e){
	    throw new IOException("Error reading .gpe file\n"+e);
	}
	return tempname;
    }

    /*
     * creates .inf and .out files from filename and writes data from PattOut
     */
    public static String unparse(String filename, GenpatExp pattOut)throws IOException{
	int lastIndex = filename.lastIndexOf(".");
	String rootname = (lastIndex==-1)?filename:filename.substring(0,lastIndex);
	int count = 1;
	String tempname=rootname;
	while(new File(tempname+".inf").exists()||new File(tempname+".out").exists()){
	    tempname = rootname+"_"+Integer.toString(count);
	    count++;
	}
	writeOut(tempname+".out",pattOut);
	writeInf(tempname+".inf",pattOut);
	return tempname;
    }

    /*
     * writes data from pattOut into a .out file
     */
    private static void writeOut(String filename, GenpatExp pattOut)throws IOException{
	new File(filename.substring(0,filename.lastIndexOf(File.separator))).mkdirs();
	BufferedWriter outbw = new BufferedWriter(new FileWriter(filename));

	outbw.write(pattOut.getTag()+"\n");
	outbw.write(pattOut.getChisq()+"\t"+pattOut.getAtomsA()+"\t"
		    +pattOut.getAtomsF()+"\t"+pattOut.getEvidence()+"\n");
	outbw.write(pattOut.getENSEMBLE()+"\t"+pattOut.getNiter()+"\t"
		    +pattOut.getNshow()+"\t"+pattOut.getIseed()+
		    ((pattOut.getSeedP()==0)?"\n":("\t"+pattOut.getSeedP()+"\n")));
	outbw.write(pattOut.getNsol()+"\t"+pattOut.getNa()+"\t"+pattOut.getNf()+"\n");
	outbw.write(pattOut.getLineMeasure()+"\t"+pattOut.getPtwidth()+"\n");
	outbw.write(pattOut.getLineMeasure()+"\t"+pattOut.getPtwidth()+"\t"+pattOut.getNf()+"\n");

	double[][] data = pattOut.getData();
	double[][] mock = pattOut.getMock();
	double[][] noise = pattOut.getNoise();
	for(int i=0;i<pattOut.getNa();i++){
	    for(int j=0;j<pattOut.getNf();j++){
		outbw.write("\t"+data[i][j]+"\t"+mock[i][j]+"\t"+noise[i][j]+"\n");
	    }
	}

	outbw.close();
    }

    /*
     * writes data from pattOut into a .inf file
     */
    private static void writeInf(String filename, GenpatExp pattOut)throws IOException{
	new File(filename.substring(0,filename.lastIndexOf(File.separator))).mkdirs();
	BufferedWriter infbw = new BufferedWriter(new FileWriter(filename));

	infbw.write(pattOut.getTag()+"\n");
	infbw.write(pattOut.getNsol()+"\t"+pattOut.getNa()+"\t"+pattOut.getNf()+"\n");
	infbw.write(pattOut.getDate()+"\n");
	infbw.write(pattOut.getNfoot()+"\n");
	
	double[] footprint = pattOut.getFootprint();
	for(int i=0;i<pattOut.getNfoot();i++){
	    infbw.write(footprint[i]+"\n");
	}

	double[][] Distribution = pattOut.getDistribution();
	double[][] DSigma = pattOut.getDistSD();
	for(int i=0;i<pattOut.getNsol();i++){
	    for(int j=0;j<pattOut.getNa();j++){
		infbw.write("\t"+Distribution[j][i]+"\t"+DSigma[j][i]+"\n");
	    }
	}

	double[][] Pattern = pattOut.getPattern();
	double[][] PSigma = pattOut.getPattSD();
	
	for(int i = 0; i < pattOut.getNsol(); i++) {
	    for(int j = 0; j < pattOut.getNf(); j++) { 
		infbw.write("\t"+Pattern[i][j]+"\t"+PSigma[i][j]+"\n");
	    }
	}

	infbw.close();
    }

    /*
     * main method takes the file name of the data files without extension as the only argument
     * parsePattOutput will create a serialized .gpe file in that directory.
     */
    public static void main(String argv[]){
	try{
	    System.out.println("wrote "+writeGPE(argv[0],parse(argv[0])));
	}catch(IOException e){
	    System.out.println(e);
	    System.exit(1);
	}
    }
}
