// Class: DataPane
//
// Purpose: A Panel which displays an object's fields and their values
//
// History: v1.1 11/06 WFS
//
// COPYRIGHT, 2006, Fox Chase Cancer Center
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the license, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have recieved a copy of the GNU General Public License
// along this this program;  if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA   02111-3107, USA.

package textfields;

import java.awt.BorderLayout;
import java.awt.Dimension;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;

import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSplitPane;
import javax.swing.JTextArea;
import javax.swing.text.AttributeSet;
import javax.swing.text.BadLocationException;
import javax.swing.text.PlainDocument;

public class DataPane extends JPanel{
    private static final long serialVersionUID = 2006011l;
    Object data;        // object whose fields are displayed
    Field[] fields;     // fields to be displayed

    int mLength = 5;    // maximum number of array elements to be displayed

    JSplitPane split;   // split pane which contains the two text areas
    JTextArea dataText; // text area for the data values
    JTextArea nameText; // text area for the names

    private boolean change = false;

    /*
     * Creates a JTextArea designed for displaying values of variables.
     * by default, it displays all of the public fields of the object.
     */
    public DataPane(Object obj){
	data = obj;
	fields = data.getClass().getFields();
	init();
	refresh();
    }

    /*
     * Constructor specifies the names of the fields to be displayed
     */
    public DataPane(Object obj, String[] f)throws NoSuchFieldException{
	data = obj;
	fields = new Field[f.length];
	for(int i=0;i<fields.length;i++){
	    fields[i]=obj.getClass().getField(f[i]);
	}
	init();
	refresh();
    }

    /*
     * Constructor specifies the actual fields to be displayed.
     * Since Field is not serializable, this cannot be used with rmi use DataPane(Object,String[])
     */
    public DataPane(Object obj, Field[] f){
	data = obj;
	fields = f;
	init();
	refresh();
    }

    /*
     * initialize the GUI
     * the gui consists of a splitpane with two text areas and scroll panes.
     * the scroll panes share a vertical scroll bar, so the data always lines up.
     * nameText cannot be edited, and dataText has a PlainDocument which restricts
     * editing to the datatypes of the associated fields.
     */
    private void init(){
	this.setLayout(new BorderLayout());

	JScrollPane nameScroll = new JScrollPane();
	nameText = new JTextArea();
	nameText.setTabSize(2);
	nameText.setEditable(false);
	nameText.setMinimumSize(new Dimension(10,20));
	nameScroll.getViewport().add(nameText);

	JScrollPane dataScroll = new JScrollPane();
	dataText = new JTextArea();
	dataText.setMinimumSize(new Dimension(10,20));
	dataText.setDocument(new PlainDocument(){
		private static final long serialVersionUID = 200601;
		public void insertString(int offs, String str, 
					 AttributeSet a)throws BadLocationException{
		    replace(offs,0,str,a);
		}
		public void remove(int offs, int len)throws BadLocationException{
		    replace(offs,len,"",null);
		}
		/*
		 * tests to make sure the number of lines is unchanged.
		 * then checks each edited line to make sure the line  still
		 * represents the correct datatype
		 */
		public void replace(int offs,int len,String text,
				    AttributeSet a)throws BadLocationException{
		    int lineAdd=text.split("\n",-1).length;
		    int lineSub=dataText.getText().substring(offs,offs+len)
			.split("\n",-1).length;
		    if(lineAdd-lineSub+dataText.getLineCount()==fields.length){
			String newText = dataText.getText().substring(0,offs)+text+
			    dataText.getText().substring(offs+len);
			String[] lines = newText.split("\n",-1);
			for(int i=0;i<lineAdd;i++){
			    int temp = dataText.getLineOfOffset(offs)+i;
			    if(!isValid(lines[temp],fields[temp])){
				return;
			    }
			}
			super.remove(offs,len);
			super.insertString(offs,text,a);
			if(change){
			    refresh();
			}
		    }
		}
	    });
	dataText.setTabSize(2);
	dataScroll.getViewport().add(dataText);
	dataScroll.setVerticalScrollBar(nameScroll.getVerticalScrollBar());

	split = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,true,nameScroll,dataScroll);
	split.setResizeWeight(0);
	split.setEnabled(true);
	split.setDividerSize(1);
	this.add(split);
    }

    /*
     * Sets the displayed fields to those specified by the names in fnames
     */
    public void setFields(String[] fnames){
	Field[] f = new Field[fnames.length];
	for(int i=0;i<f.length;i++){
	    try{
		f[i] = data.getClass().getField(fnames[i]);
	    }catch(NoSuchFieldException nsfe){
		System.out.println(nsfe);
	    }
	}
	fields = f;
	refresh();
    }

    /*
     * Sets the displayed fields to f
     * Will not work with rmi because Field is not serializable.
     * use setFields(String[])
     */
    public void setFields(Field[] f){
	fields = f;
	refresh();
    }

    /*
     * returns the displayed fields
     */
    public Field[] getFields(){
	return fields;
    }

    /*
     * wrapper around split.setDividerLocation(double)
     */
    public void setDividerLocation(double proportionalLocation){
	split.setDividerLocation(proportionalLocation);
    }

    /*
     *wrapper around split.setDividerLocation(int)
     */
    public void setDividerLocation(int location){
	split.setDividerLocation(location);
    }

    /*
     * sets the object whose fields are displayed
     */
    public void setData(Object obj){
	data = obj;
	refresh();
    }

    /*
     * returns the object whose fields are displayed
     */
    public Object getData(){
	return data;
    }

    /*
     * erases the data Object and clears the display
     */
    public void clear(){
	data=new Object();
	refresh();
    }

    /*
     * refreshes the display. This is calles whenever the data is changed
     * by a method and should be called whenever a program changes the data
     */
    public void refresh(){
	int l = 0;
	int p = 0;
	try{//gets the previous caret position so that it can be replaced after refresh
	    p = dataText.getCaretPosition();
	    p -=dataText.getLineStartOffset(dataText.getLineOfOffset(p));
	    l = dataText.getLineOfOffset(dataText.getCaretPosition());
	}catch(Exception e){}

	String t1 = "";
	String t2 = "";

	for(int i=0;i<fields.length;i++){
	    t1+=fields[i].getName()+((i==(fields.length-1))?"":"\n");
	    try{
		t2+=printArray(fields[i].get(data),mLength);
	    }catch(NullPointerException npe){
	    }catch(IllegalAccessException iae){}
	    t2+=((i==(fields.length-1))?"":"\n");
	}

	nameText.setText(t1);
	dataText.setText(t2);

	try{//replaces the caret to the original position
	    dataText.setCaretPosition(Math.min(dataText.getLineStartOffset(l)+p,
							   dataText.getLineEndOffset(l)-1));
	}catch(Exception e){}
    }

    /*
     * sets the maximum number of elements of an array to be displayed
     */
    public void setmaxLength(int newMax){
	mLength = newMax;
	refresh();
    }

    /*
     * used to determine whether or not the change by the user should be allowed
     * this tries to store the new text value as the data and, if it fails, does
     * not allow the change. If the string becomes "", the value is set to 0
     */
    private boolean isValid(String line, Field f){
	change = false;
	try{
	    Constructor c = f.get(data).getClass().getConstructor(new Class[]{line.getClass()});
	    f.set(data,c.newInstance(new Object[]{line}));
	    change = ((pattviewer.PattInput)data).constrain(f.getName());
	}catch(NoSuchMethodException nsme){
	    try{
		return parseArray(line,f.get(data),mLength);
	    }catch(IllegalAccessException iae){
		return false;
	    }
	}catch(Exception e){
	    try{
		if(line.equals("")){
		    Constructor c
			=f.get(data).getClass().getConstructor(new Class[]{"0".getClass()});
		    f.set(data,c.newInstance(new Object[]{"0"}));
		    change = ((pattviewer.PattInput)data).constrain(f.getName());
		    return true;
		}
	    }catch(Exception e2){}
	    try{
		return line.equals(f.get(data).toString());
	    }catch(Exception e3){}
	    return false;
	}
	return true;
    }

    /*
     * returns a string representation of the elements of the array up to number maxLength
     * simply calls the toString method if o is not an array
     */
    private static String printArray(Object o, int maxLength){
	if(o.getClass().isArray()){
	    String output = "[\t";
	    int al = 0;
	    if(new Object[]{}.getClass().isAssignableFrom(o.getClass())){
		Object[] array = ((Object[])o);
		al = array.length;
		int l = Math.min(maxLength,al);
		for(int i=0;i<l;i++){
		    output+=printArray(array[i], maxLength)+"\t";
		}
	    }else if(new double[]{}.getClass().isAssignableFrom(o.getClass())){
		double[] array = ((double[])o);
		al = array.length;
		int l = Math.min(maxLength,al);
		for(int i=0;i<l;i++){
		    output+=array[i]+"\t";
		}
	    }else if(new int[]{}.getClass().isAssignableFrom(o.getClass())){
		int[] array = ((int[])o);
		al = array.length;
		int l = Math.min(maxLength,al);
		for(int i=0;i<l;i++){
		    output+=array[i]+"\t";
		}
	    }else if(new long[]{}.getClass().isAssignableFrom(o.getClass())){
		long[] array = ((long[])o);
		al = array.length;
		long l = Math.min(maxLength,al);
		for(int i=0;i<l;i++){
		    output+=array[i]+"\t";
		}
	    }
	    if(al>maxLength){
		output+="...\t";
	    }
	    output+="]";
	    return output;
	}
	if(o.toString().equals("0")){
	    return "";
	}
	return o.toString();
    }

    /*
     * splits the String representing an array into a String[] containing its elements
     * this is an extension of String.split meant to recognize arrays as elements
     * e.g. [ [ 1 2 3 ] [ 4 5 ] 6 [ [ 7 ] 8 ] ]
     *      becomes [
     *              [ 1 2 3 ]
     *              [ 4 5 ]
     *              6
     *              [ [ 7 ] 8 ] 
     *              ]
     */
    private static String[] split(String line){
	String[] parts = line.split("\t");
	String temp = "";
	int counter = 0;
	int count = 0;
	for(int i=0;i<parts.length;i++){
	    temp += parts[i]+"\t";
	    counter += (parts[i].equals("[")?1:0)-(parts[i].equals("]")?1:0);
	    if(counter<=1){
		parts[count]=temp.trim();
		temp = "";
		count++;
	    }
	}
	String[] output = new String[count];
	for(int i=0;i<count;i++){
	    output[i]=parts[i];
	}
	return output;
    }

    /*
     * determines whether a string is a valid representation of the array data and stores the
     * data if it is. This method recursively calls itself if the string contains arrays as
     * elements.
     */
    private static boolean parseArray(String s, Object o, int maxLength){
	if(o.getClass().isArray()){
	    String[] e = split(s);
	    try{
		int al = 0;
		int l = 0;
		if(new Object[]{}.getClass().isAssignableFrom(o.getClass())){ // object array
		    Object[] array = (Object[])o;
		    al = array.length;
		    l = Math.min(maxLength,al);
		    for(int i=0;i<l;i++){
			if(array[i].getClass().isArray()){
			    if(!parseArray(e[i+1],array[i],maxLength)){
				return false;
			    }
			}else if(e[i+1].equals("")){
			    Constructor c=array[i].getClass().getConstructor(new Class[]{"0".getClass()});
			    array[i]=c.newInstance(new Object[]{"0"});
			}else{
			    Constructor c=array[i].getClass().getConstructor(new Class[]{e[i+1].getClass()});
			    array[i]=c.newInstance(new Object[]{e[i+1]});
			}
		    }
		}else if(new double[]{}.getClass().isAssignableFrom(o.getClass())){ // double array
		    double[] array = (double[])o;
		    al = array.length;
		    l = Math.min(maxLength,al);
		    for(int i=0;i<l;i++){
			if(e[i+1].equals("")){
			    array[i]=Double.parseDouble("0");
			}else{
			    array[i]=Double.parseDouble(e[i+1]);
			}
		    }
		}else if(new int[]{}.getClass().isAssignableFrom(o.getClass())){ // int array
		    int[] array = (int[])o;
		    al = array.length;
		    l = Math.min(maxLength,al);
		    for(int i=0;i<l;i++){
			if(e[i+1].equals("")){
			    array[i]=Integer.parseInt("0");
			}else{
			    array[i]=Integer.parseInt(e[i+1]);
			}
		    }
		}else if(new long[]{}.getClass().isAssignableFrom(o.getClass())){ // long array
		    long[] array = (long[])o;
		    al = array.length;
		    l = Math.min(maxLength,al);
		    for(int i=0;i<l;i++){
			if(e[i+1].equals("")){
			    array[i]=Long.parseLong("0");
			}else{
			    array[i]=Long.parseLong(e[i+1]);
			}
		    }
		}
		return !(!e[0].equals("[")||!e[e.length-1].equals("]")||
			 ((l<al)?l!=e.length-3||!e[e.length-2].equals("..."):l!=e.length-2));
	    }catch(Exception ex){}
	}
	return s.equals(o.toString());
    }
}
